#pragma once

#include <pcl\point_types.h>
#include <pcl\point_cloud.h>
#include <pcl\segmentation\sac_segmentation.h>
#include "data_def_api.h"
#include "data_def.h"
#include "logger.h"
#include "imstorage_manager.h"

namespace graft_cv {
	class CChessboardCalibration {
	public:
		CChessboardCalibration(ConfigParam&c, CGcvLogger*pLog = 0);
		~CChessboardCalibration();
		int cross_detect(PositionInfo& posinfo);
		float* get_raw_point_cloud(int &data_size);
		int load_data(float*pPoint, int pixel_size, int pt_size, int dtype, const char* fn = 0);
		void set_image_saver(CImStoreManager** ppis) { m_ppImgSaver = ppis; }
	private:
		//global configure parameters
		ConfigParam& m_cparam;
		CGcvLogger * m_pLogger;
		int m_dtype;
		string m_pcdId;

		CImStoreManager** m_ppImgSaver;

		pcl::PointCloud<pcl::PointXYZ>::Ptr m_raw_cloud;
		double m_cloud_mean_dist;

		double min_dist_of_lines(
			pcl::ModelCoefficients::Ptr line_1,//input
			pcl::ModelCoefficients::Ptr line_2//input
			);
		void virtual_cross_of_lines(
			pcl::ModelCoefficients::Ptr line_1, //input
			pcl::ModelCoefficients::Ptr line_2,//input
			pcl::PointXYZ& cross_pt//output
		);


		int read_ply_file(const char* fn);
		double compute_nearest_neighbor_distance(pcl::PointCloud<pcl::PointXYZ>::Ptr);
		void cloud_mean_dist(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,	//input 
			double& mean_dist
		);
		void gen_result_img(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud_raw,//룬δ˵in_cloud_raw, 
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud_edge,	//룬Ե, 
			std::vector<pcl::PointXYZ>& cross_points,				//룬,		
			cv::Mat& rst_img						//ͼƬ 1280*1280
		);
		

		void viewer_cloud(pcl::PointCloud<pcl::PointXYZ>::Ptr, std::string&winname);
		void viewer_cloud(pcl::PointCloud<pcl::PointXYZRGB>::Ptr, std::string&winname);

		void viewer_cloud_debug(
			pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud,
			pcl::PointXYZ&p,//ץȡ
			pcl::PointXYZ &p_ref,//ֲ
			pcl::PointXYZ &root_pt,
			std::string&winname);

		void viewer_cloud_cluster(pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, std::vector<pcl::PointXYZ>cluster_center, std::string&winname);
		void viewer_cloud_cluster_box(
			pcl::PointCloud<pcl::PointXYZ>::Ptr cloud,
			std::vector<pcl::PointXYZ>&cluster_center,
			std::vector<pcl::PointXYZ>&cluster_box,
			std::vector<std::vector<int> >& clt_line_idx,
			std::string&winname);
	};

};