#pragma once

#include <vector>
#include <string>
#include <math.h>
#include <pcl\point_types.h>
#include <pcl\point_cloud.h>
#include "logger.h"

namespace graft_cv {
	class CStemResult {
	public:
		CStemResult() :
			root_x(0.0),
			root_z(0.0),
			stem_size(0),
			batch_id(std::string("")),
			root_count(0)			
		{};
		CStemResult(double x,double y, double z, int size, std::string& bid, int rcount)
		{
			
			root_x = x;
			root_y = y;
			root_z = z;
			stem_size = size;
			batch_id = std::string(bid);
			root_count = rcount;
		};
		double calcluate_distance(const CStemResult& sr) {
			double dist = 0.0;
			dist = sqrt((this->root_x - sr.root_x) *(this->root_x - sr.root_x) +
				//(this->root_y - sr.root_y) *(this->root_y - sr.root_y) +
				(this->root_z - sr.root_z) *(this->root_z - sr.root_z));
			return dist;
		};
		double root_x;	//ʶ𵽵ľx
		double root_y;
		double root_z;	//ʶ𵽵ľz
		int stem_size;	//ʶ𵽵ľ
		std::string batch_id;	//ʶ𵽵ľΣÿһδһ£
		int root_count;	//ͳrootۼƸ

	private:
	};

	class CStemResultInfos {
	public:
		CStemResultInfos(
			double seedling_dist,
			int holes_number,
			double x_min,
			double x_max,
			double z_min,
			double z_max,
			std::string pcd_id,
			CGcvLogger*pLog=0);
		~CStemResultInfos();
		
		//void set_json_filename(std::string& filename);
		//void read_root_centers(std::string& filename);//ȡʷݣÿʱȥȡ
		//void write_root_centers(std::string& filename);//浱ǰݣһʱ䱣һ

		void append(CStemResult& sr);
		void clear();
		void get_root_centers(std::vector<CStemResult>&rst);


	protected:
		CGcvLogger * m_pLogger;
		int m_max_size;	//m_infos󳤶		
		std::vector<CStemResult> m_infos;
		double m_seedling_dist;		//ֲ࣬
		std::vector<CStemResult> m_root_centers;	//λʷƽֵ
		//int m_append_counter;
		//std::string m_json_filename;
		std::string m_pcdId;
		int m_holes_number;
		double m_xmin;
		double m_xmax;
		double m_zmin;
		double m_zmax;

		void _update_root_centers(CStemResult& sr);	//ͨǰɾλã
		//void _filter_root_centers(double d1, double d2);	//ɵĸm_root_centersйˣ޳쳣
		//void euclidean_clustering_ttsas(
		//	pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
		//	double d1, double d2,
		//	std::vector<pcl::PointXYZ>&cluster_center,
		//	std::vector<std::vector<int>> &clustr_member
		//);
		void gen_root_centers();
	};

	//ֲ״̬¼ֲƷֲȷֲǷȡߣѨλǷ
	class CSeedlingStatus {
	public:
		CSeedlingStatus(
			int dtype,		//1-ľ 0-
			double step,	//xֱͼƷֲÿbinĿȣĬ5ף
			double x_min,	//x
			double x_max,	//x
			double pc_mean_dist,	//ƣƽ
			CGcvLogger*pLog = 0);
		~CSeedlingStatus();
		void set_x_centers(std::vector<double>&cx);
		void append_hist(std::vector<int>&xhist,	//input
			std::vector<bool>&xstatus);				//output
		
	protected:
		CGcvLogger * m_pLogger;
		int m_max_size;	//¼
		int m_record_cursor;
		cv::Mat m_history_histogram;	//ʷxֱֲͼ
		cv::Mat m_history_point_size;	//ʷƿ
		cv::Mat m_history_status;		//ʷѨλ״̬
		std::vector<int> m_idx_low;		//ѨλӦŵλ
		std::vector<int> m_idx_up;		//ѨλӦŵλ

		int m_dtype;					//1-ľ 0-
		double m_xmin;					//x
		double m_xmax;					//x
		double m_bin_step;				//xֱͼƷֲÿbinĿȣĬ5ף
		double m_pc_mean_dist;			//ƣƽ
		std::vector<double>m_center_x;
		int m_hist_length;

		void get_status(std::vector<bool>&xstatus);
		
		
	};
}